/* /////////////////////////////////////////////////////////////////////////////////////////////////
ParentHoldLink v2 by MacGruber.
The build-in ParentLink feature in VaM that can link one atom to another has the problem of using the position the atom was at in the previous frame. Even at 90fps, for fast animations this can cause obvious visual lag. ParentHoldLink solves this issue, updating the atom to the correct position. This plugin is based on the identically named plugin by Blazedust, but has some improvements.

Version 2 - 2021-02-22
	Added "Reset when disabled" toggle.

Version 1 - 2020-10-05
	Saving link offset with the scene to avoid misalignment when loading the scene later.
	Handling plugin enable/disable.
	Code cleanup and minor performance improvements.

Original plugin by Blazedust - 2020-06-07
	https://hub.virtamate.com/resources/parentholdlink-for-customunityassets.415/

///////////////////////////////////////////////////////////////////////////////////////////////// */


using System;
using UnityEngine;
using UnityEditor;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;

namespace MacGruber
{
	public class ParentHoldLink : MVRScript
	{
		private JSONStorableBool myResetWhenDisabled;
		private Vector3 myLinkPos;
		private Quaternion myLinkRot;
		private bool myPosInitDone = false;
		private bool myRotInitDone = false;

		public override void Init()
		{
			SetupInfoText(
				"<color=#606060><size=40><b>ParentHoldLink v0.2</b></size>\nThe build-in <b><i>ParentLink</i></b> feature in VaM that can link one atom to another has the problem of using the position the atom was at in the previous frame. Even at 90fps, for fast animations this can cause obvious visual lag. <b><i>ParentHoldLink</i></b> solves this issue, updating the atom to the correct position. This plugin is based on the identically named plugin by Blazedust, but has some improvements.</color>",
				500.0f, false
			);
			
			myResetWhenDisabled = SetupToggle("Reset when disabled", true, false);

			SetupInfoText(
				"<b>Usage:</b>\n" +
				"1. Add this plugin to child atom you want to attach to some parent object.\n" +
				"2. Go to the <b><i>Control</i></b> tab of the child atom.\n" +
				"3. Set <b><i>Link To Atom</i></b> and <b><i>Link To</i></b> to the parent object. Alternatively use the <b><i>Select Link To From Scene</i></b> button.\n" +
				"4. Set the <b><i>Hold</i></b> checkbox for <b><i>Position</i></b> and/or <b><i>Rotation</i></b>.\n" +
				"5. You can grab the child atom as usual to adjust its relative position.\n" +
				"6. Disable <b>Reset when disabled</b> when you want to enabled/disable the plugin or atom via triggers and want to prevent missalignments caused by the plugin resetting it's link.",
				565.0f, true
			);
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if ((includePhysical || forceStore) && (myPosInitDone || myRotInitDone))
			{
				needsStore = true;
				if (myPosInitDone)
				{
					JSONClass pos = new JSONClass();
					pos["x"].AsFloat = myLinkPos.x;
					pos["y"].AsFloat = myLinkPos.y;
					pos["z"].AsFloat = myLinkPos.z;
					jc["position"] = pos;
				}
				if (myRotInitDone)
				{
					JSONClass rot = new JSONClass();
					Vector3 euler = myLinkRot.eulerAngles;
					rot["x"].AsFloat = euler.x;
					rot["y"].AsFloat = euler.y;
					rot["z"].AsFloat = euler.z;
					jc["rotation"] = rot;
				}
			}
			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!physicalLocked && restorePhysical)
			{
				if (jc.HasKey("position"))
				{
					JSONClass pos = jc["position"].AsObject;
					myLinkPos.x = pos["x"].AsFloat;
					myLinkPos.y = pos["y"].AsFloat;
					myLinkPos.z = pos["z"].AsFloat;
					myPosInitDone = true;
				}
				if (jc.HasKey("rotation"))
				{
					JSONClass rot = jc["rotation"].AsObject;
					Vector3 euler;
					euler.x = rot["x"].AsFloat;
					euler.y = rot["y"].AsFloat;
					euler.z = rot["z"].AsFloat;
					myLinkRot.eulerAngles = euler;
					myRotInitDone = true;
				}
			}
		}

 		private void OnEnable()
		{
			if (myResetWhenDisabled.val)
			{
				myPosInitDone = false;
				myRotInitDone = false;
			}
		}

		private void OnDisable()
		{
			if (myResetWhenDisabled.val)
			{
				myPosInitDone = false;
				myRotInitDone = false;
			}
		}

		private void LateUpdate()
		{
			FreeControllerV3 mainController = GetMainController();
			if (mainController.linkToRB != null)
			{
				Transform linkTransform = mainController.linkToRB.transform;
				if (mainController.currentPositionState != FreeControllerV3.PositionState.Hold)
				{
					myPosInitDone = false;
				}
				else if (!myPosInitDone)
				{
					myLinkPos = linkTransform.InverseTransformPoint(containingAtom.childAtomContainer.position);
					myPosInitDone = true;
				}

				if (mainController.currentRotationState != FreeControllerV3.RotationState.Hold)
				{
					myRotInitDone = false;
				}
				else if (!myRotInitDone)
				{
					myLinkRot = Quaternion.Inverse(linkTransform.rotation) * containingAtom.childAtomContainer.rotation;
					myRotInitDone = true;
				}

				if (myPosInitDone || myRotInitDone)
				{
					Vector3 pos = myPosInitDone ? linkTransform.TransformPoint(myLinkPos) : mainController.transform.position;
					Quaternion rot = myRotInitDone ? linkTransform.rotation * myLinkRot : mainController.transform.rotation;
					mainController.transform.SetPositionAndRotation(pos, rot);
				}
			}
			else
			{
				myPosInitDone = false;
				myRotInitDone = false;
			}
		}

		private JSONStorableString SetupInfoText(string text, float height, bool rightSide)
		{
			JSONStorableString storable = new JSONStorableString("Info", text);
			UIDynamic textfield = CreateTextField(storable, rightSide);
			textfield.height = height;
			return storable;
		}
		
		private JSONStorableBool SetupToggle(string label, bool defaultValue, bool rightSide)
		{
			JSONStorableBool storable = new JSONStorableBool(label, defaultValue);
			storable.storeType = JSONStorableParam.StoreType.Full;
			CreateToggle(storable, rightSide);
			RegisterBool(storable);
			return storable;
		}
	}
}