﻿/* /////////////////////////////////////////////////////////////////////////////////////////////////
PowerHandles v0.4 by MacGruber.
Improved Position/Rotation handles for VaM.

Version 0.4 2021-12-28
	Fixed issue when no longer cycling through controllers by clicking.

Version 0.3 2021-12-21
	Fixed issue when releasing modifier key (CTRL or ALT) before releasing the mouse button.
	Fixed issue with handle position/orientation not being updated when the same is set from AcidBubbles Timeline.
	Prevent switch rotation pivot (SHIFT) while actively moving/rotating the control.

Version 0.2 2021-07-17
	Fixed handles not working when "Only Show Targets and Overlays When Menu Open" is disabled in UserPreferences and the UI is closed.

Version 0.1 2021-01-17
	Initial release.

///////////////////////////////////////////////////////////////////////////////////////////////// */

using UnityEngine;
using Battlehub.RTCommon;
using GridMode = FreeControllerV3.GridMode;
using System.Collections.Generic;

using System.Text;
using UnityEngine.UI;
using UnityEngine.Events;

namespace MacGruber
{
    public class PowerHandles : MVRScript
    {
		private Transform myMouseGrab;
		private FreeControllerV3 mySelectedController;
		private FreeControllerV3 myActiveController;
		private GridMode myPositionGridMode = GridMode.None;
		private GridMode myRotationGridMode = GridMode.None;
		private float myPositionGridAlign;
		private float myRotationGridAlign;

		private JSONStorableFloat myPositionGridAlignUI;
		private JSONStorableFloat myRotationGridAlignUI;
		private JSONStorableFloat myGizmoScaleUI;

		public override void Init()
		{
			SetupInfoText(
				"<color=#606060><size=40><b>PowerHandles v0.4</b></size>\nImproved Position/Rotation handles for VaM Desktop Mode.</color>\n\n" +
				"<b>Usage:</b>\nLeftCTRL:  Enable Position\nLeftALT:     Enable Rotation\nLeftSHIFT: Global Coordinates\nSPACE:     Enable Grid Mode",
				350.0f, true
			);

			myGizmoScaleUI = SetupSliderFloat("Gizmo Scale", 1.0f, 0.05f, 1.0f, false);
			myGizmoScaleUI.setCallbackFunction += SyncMeshScale;
			SyncMeshScale(myGizmoScaleUI.val);

			myPositionGridAlignUI = SetupSliderFloat("Position Grid", 0.1f, 0.01f, 1.0f, false);
			myPositionGridAlignUI.setCallbackFunction += (float v) => {
				v = Mathf.Round(v*100.0f)/100.0f;
				myPositionGridAlignUI.valNoCallback = v;
				if (mySelectedController != null)
					mySelectedController.positionGrid = v;
			};

			myRotationGridAlignUI = SetupSliderInt("Rotation Grid", 15, 1, 90, false);
			myRotationGridAlignUI.setCallbackFunction += (float v) => {
				if (mySelectedController != null)
					mySelectedController.rotationGrid = v;
			};
		}

		private void Update()
		{
			SuperController sc = SuperController.singleton;
			UserPreferences up = UserPreferences.singleton;
			bool handlesEnabled = sc.MonitorCenterCamera != null && (!up.showTargetsMenuOnly || sc.mainHUD.gameObject.activeSelf);
			bool usePosition = handlesEnabled && Input.GetKey(KeyCode.LeftControl);
			bool useRotation = handlesEnabled && Input.GetKey(KeyCode.LeftAlt);
			bool useGlobal = handlesEnabled && Input.GetKey(KeyCode.LeftShift);
			bool useGrid = handlesEnabled && Input.GetKey(KeyCode.Space);

			if (myActiveController != null && Input.GetMouseButtonUp(0))
			{
				// force select controller in case modifer key was relased before mouse button
				sc.SelectController(myActiveController, openUI: false);
				myActiveController = null;
			}

			FCPositionHandle positionHandle = sc.selectedControllerPositionHandle;
			FCRotationHandle rotationHandle = sc.selectedControllerRotationHandle;
			if (positionHandle != null)
			{
				if (usePosition)
				{
					myActiveController = positionHandle.controller;
				}
				else if (positionHandle.enabled) // modifer was released, but mouse button might still be down => ForceDrop
				{
					positionHandle.ForceDeselectAxis();
					positionHandle.ForceDrop();
				}
				else if (positionHandle.controller != null) // sync position manually as positionHandle is not updated
				{
					positionHandle.transform.position = positionHandle.controller.transform.position;
				}

				positionHandle.enabled = usePosition;

				if (positionHandle.IsDragging) // Prevent switch pivot while active
					useGlobal = RuntimeTools.PivotRotation == RuntimePivotRotation.Global;
			}
			if (rotationHandle != null)
			{
				if (useRotation)
				{
					myActiveController = rotationHandle.controller;
				}
				else if (rotationHandle.enabled)  // modifer was released, but mouse button might still be down => ForceDrop
				{
					rotationHandle.ForceDeselectAxis();
					rotationHandle.ForceDrop();
				}
				else if (rotationHandle.controller != null)  // sync rotation manually as rotationHandle is not updated
				{
					rotationHandle.transform.rotation = rotationHandle.controller.transform.rotation;
				}

				rotationHandle.enabled = useRotation;

				if (rotationHandle.IsDragging) // Prevent switch pivot while active
					useGlobal = RuntimeTools.PivotRotation == RuntimePivotRotation.Global;
			}

			RuntimeTools.PivotRotation = useGlobal ? RuntimePivotRotation.Global : RuntimePivotRotation.Local;

			FreeControllerV3 controller = sc.GetSelectedController();
			if (controller != mySelectedController)
			{
				SelectController(controller);
			}
			if (mySelectedController != null)
			{
				GridMode gridMode = GridMode.None;
				if (useGrid)
					gridMode = useGlobal ? GridMode.Global : GridMode.Local;
				if (mySelectedController.positionGridMode != gridMode)
					mySelectedController.positionGridMode = gridMode;
				if (mySelectedController.rotationGridMode != gridMode)
					mySelectedController.rotationGridMode = gridMode;
			}
		}

		private void OnAtomAdded(Atom atom)
		{
			if (myGizmoScaleUI != null)
				SyncMeshScale(myGizmoScaleUI.val);
		}

		private void OnEnable()
		{
			SuperController sc = SuperController.singleton;

			myMouseGrab = sc.mouseGrab;
			sc.mouseGrab = null;

			sc.onAtomAddedHandlers += OnAtomAdded;

			if (myGizmoScaleUI != null)
				SyncMeshScale(myGizmoScaleUI.val);
		}

		private void OnDisable()
		{
			ResetController();

			SuperController sc = SuperController.singleton;
			bool handlesEnabled = sc.mainHUD.gameObject.activeSelf && sc.MonitorCenterCamera != null;

			FCPositionHandle positionHandle = sc.selectedControllerPositionHandle;
			FCRotationHandle rotationHandle = sc.selectedControllerRotationHandle;
			if (positionHandle != null)
			{
				positionHandle.ForceDeselectAxis();
				positionHandle.ForceDrop();
				positionHandle.enabled = handlesEnabled;
			}
			if (rotationHandle != null)
			{
				rotationHandle.ForceDeselectAxis();
				rotationHandle.ForceDrop();
				rotationHandle.enabled = handlesEnabled;
			}

			RuntimeTools.PivotRotation = RuntimePivotRotation.Local;

			sc.mouseGrab = myMouseGrab;
			myMouseGrab = null;

			sc.onAtomAddedHandlers -= OnAtomAdded;
			SyncMeshScale(1.0f);
		}

		private void SelectController(FreeControllerV3 controller)
		{
			ResetController();
			mySelectedController = controller;
			if (mySelectedController != null)
			{
				myPositionGridMode = mySelectedController.positionGridMode;
				myRotationGridMode = mySelectedController.rotationGridMode;
				myPositionGridAlign = mySelectedController.positionGrid;
				myRotationGridAlign = mySelectedController.rotationGrid;
				mySelectedController.positionGrid = myPositionGridAlignUI.val;
				mySelectedController.rotationGrid = myRotationGridAlignUI.val;

				// Hack to hide collider mesh of Empty atoms when collision is disabled
				if (mySelectedController.containingAtom.type == "Empty" && mySelectedController.containingAtom.collisionEnabledJSON != null)
				{
					if (!mySelectedController.containingAtom.collisionEnabled)
						SyncCollisionEnabled(false);
					mySelectedController.containingAtom.collisionEnabledJSON.setCallbackFunction += SyncCollisionEnabled;
				}
			}
		}

		private void ResetController()
		{
			if (mySelectedController != null)
			{
				mySelectedController.positionGridMode = myPositionGridMode;
				mySelectedController.rotationGridMode = myRotationGridMode;
				mySelectedController.positionGrid = myPositionGridAlign;
				mySelectedController.rotationGrid = myRotationGridAlign;

				if (mySelectedController.containingAtom.type == "Empty" && mySelectedController.containingAtom.collisionEnabledJSON != null)
				{
					mySelectedController.containingAtom.collisionEnabledJSON.setCallbackFunction -= SyncCollisionEnabled;
				}

				mySelectedController = null;
			}
		}

		private void SyncCollisionEnabled(bool v)
		{
			Transform t = mySelectedController?.containingAtom?.reParentObject?.Find("object/rescaleObject/SphereRender");
			if (t != null)
				t.gameObject.SetActive(v);
		}

		private void SyncMeshScale(float scale)
		{
			List<FreeControllerV3> controllers = SuperController.singleton.GetAllFreeControllers();
			for (int i=0; i<controllers.Count; ++i)
			{
				FreeControllerV3 c = controllers[i];
				c.deselectedMeshScale = 0.02f * scale;
				c.meshScale = 0.60f * scale;
			}
		}

		private JSONStorableFloat SetupSliderFloat(string name, float defaultValue, float minValue, float maxValue, bool rightSide)
		{
			JSONStorableFloat storable = new JSONStorableFloat(name, defaultValue, minValue, maxValue, true, true);
			storable.storeType = JSONStorableParam.StoreType.Full;
			CreateSlider(storable, rightSide);
			RegisterFloat(storable);
			return storable;
		}

		private JSONStorableFloat SetupSliderInt(string label, int defaultValue, int minValue, int maxValue, bool rightSide)
		{
			JSONStorableFloat storable = new JSONStorableFloat(label, defaultValue, minValue, maxValue, true, true);
			storable.storeType = JSONStorableParam.StoreType.Full;
			UIDynamicSlider slider = CreateSlider(storable, rightSide);
			slider.slider.wholeNumbers = true;
			slider.valueFormat = "F0";
			RegisterFloat(storable);
			return storable;
		}

		private JSONStorableString SetupInfoText(string text, float height, bool rightSide)
		{
			JSONStorableString storable = new JSONStorableString("Info", text);
			UIDynamic textfield = CreateTextField(storable, rightSide);
			textfield.height = height;
			return storable;
		}
    }
}
