﻿/* /////////////////////////////////////////////////////////////////////////////////////////////////
SkyMagic Loader v0.5 by MacGruber.
Retrieves a cubemap texture from a ReflectionProbe inside your CustomUnityAsset and sets it as source for VaM's Global Sky and Global Illumination.

Version 0.5 2021-12-21
	Fixed massive performance issue related to SkyMagic used inside SubScenes.
	Fixed issue with native ReflectionProbes not having their intensity value applied when inside SubScenes.

Version 0.4 2021-01-16
	Fixed issue with SkyMagic inside SubScenes.

Version 0.3 2020-10-10
	Fixed issue where sliders in 'Scene Lighting' panel were not being applied to custom skies.
	Fixed NullReferenceException exception when loading an identical scene again

Version 0.2 2020-05-01
	Added preview sphere feature
	Removed rotation-sync feature, since ReflectionProbes do not support rotation and it was pointless. Added a warning instead.
	Fixed silent crash on cleanup.

Version 0.1 2020-04-12
	Initial release.

///////////////////////////////////////////////////////////////////////////////////////////////// */

using UnityEngine;
using UnityEngine.Events;
using System.Collections;
using mset;
using System;

namespace MacGruber
{
    public class SkyMagicLoader : MVRScript
    {
		private readonly string SKY_GAMEOBJECT_NAME = "SkyMagic Custom Sky";

		private JSONStorableFloat myIntensity;
		private JSONStorableBool myUseAsSky;
		private JSONStorableBool mySkyIsHDR;
		private JSONStorableBool mySkyIsSRGB;
		private JSONStorableBool myShowPreviewSphere;
		private bool myNeedUpdate = true;
		private bool myHasSetSky = false;

		private Sky mySky;
		private ReflectionProbe mySkyProbe;
		private GameObject myPreviewSphere;
		private Material myPreviewMaterial;

		public override void Init()
		{
			if (containingAtom.type != "CustomUnityAsset")
			{
				SuperController.LogError("The MacGruber.SkyMagic plugin needs to be placed on a CustomUnityAsset atom.");
				SetupInfoText(this,
					"<color=#606060><size=40><b>SkyMagic Loader v0.5</b></size>\n</color><color=#FF0000>SkyMagic plugin needs to be placed on a CustomUnityAsset atom.</color>\n\n",
					1200.0f, true
				);
				myNeedUpdate = false;
				return;
			}

			SetupInfoText(this,
				"<color=#606060><size=40><b>SkyMagic Loader v0.5</b></size>\nRetrieves a cubemap texture from a ReflectionProbe inside your CustomUnityAsset and sets it as source for VaM's Global Sky and Global Illumination.</color>\n\n" +
				"<color=#FF0000><b>Note that Unity ReflectionProbes do NOT support rotation! If you have non-zero rotation in this atom, your CustomUnityAsset or rotate the Global Sky, your reflections might not match the world anymore.</b></color>\n\n" +
				"<b>ReflectionProbe Intensity:</b> Sets the intensity of <b>ALL</b> ReflectionProbes inside your CustomUnityAsset. This affects all nearby Unity assets.\n\n" +
				"<b>Global Sky Enabled:</b> Uses the <b>FIRST</b> ReflectionProbe found inside your CustomUnityAsset as source for Global Sky and Global Illumination, to be used on all VaM native assets.\n\n" +
				"<b>Global Sky is sRGB:</b> Uses gamma colorspace instead of linear. Try to toggle this if your sky is too dark or too bright.\n\n" +
				"<b>Show Preview Sphere:</b> Spawns a prefectly relecting sphere.\n\n",
				1200.0f, true
			);

			myIntensity = SetupSliderFloat(this, "ReflectionProbe Intensity", 1.0f, 0.0f, 10.0f, false);
			myIntensity.setCallbackFunction += (float v) => { UpdateProbes(); };

			myUseAsSky = SetupToggle(this, "Global Sky Enabled", false, false);
			myUseAsSky.setCallbackFunction += (bool v) => { UpdateSky(v); };

			mySkyIsSRGB = SetupToggle(this, "Global Sky is sRGB", true, false);
			mySkyIsSRGB.setCallbackFunction += (bool v) => { UpdateSky(); };

			myShowPreviewSphere = SetupToggle(this, "Show Preview Sphere", false, false);
			myShowPreviewSphere.setCallbackFunction += UpdatePreview;

			SetupButton(this, "Force Sync", () => { myNeedUpdate = true; }, false);

			myNeedUpdate = true;
		}

		private void OnEnable()
		{
			myNeedUpdate = true;
		}

		private void OnDisable()
		{
			try {
			UpdateSky(false);
			GameObject skyGO = transform.Find(SKY_GAMEOBJECT_NAME)?.gameObject;
			if (skyGO != null)
				Destroy(skyGO);

			if (myPreviewMaterial != null)
				Destroy(myPreviewMaterial);
			if (myPreviewSphere != null)
				Destroy(myPreviewSphere);
			myPreviewSphere = null;
			myPreviewMaterial = null;

			}
			catch (Exception e)
			{
				SuperController.LogError(e.ToString());
			}
		}

		private void Update()
		{
			if (mySkyProbe == null && myHasSetSky)
				myNeedUpdate = true;

			if (myNeedUpdate)
			{
				bool haveProbes = UpdateProbes();
				bool haveSky = UpdateSky();
				UpdatePreview(myShowPreviewSphere.val);
				myNeedUpdate = !haveProbes || !haveSky;
			}
		}

		private bool UpdateSky()
		{
			return UpdateSky(myUseAsSky.val);
		}

		private bool UpdateSky(bool useAsSky)
		{
			SkyshopLightController sslc = SkyshopLightController.singleton;
			Transform t = containingAtom?.reParentObject?.Find("object/rescaleObject");
			if (t != null && t.childCount > 0 && useAsSky)
			{
				mySkyProbe = t.GetComponentInChildren<ReflectionProbe>();
				if (mySkyProbe != null)
				{
					GameObject skyGO = transform.Find(SKY_GAMEOBJECT_NAME)?.gameObject;
					if (skyGO == null)
					{
						skyGO = Instantiate(sslc.skies[0].gameObject, transform);
						skyGO.name = SKY_GAMEOBJECT_NAME;
					}
					mySky = skyGO.GetComponent<Sky>();
					mySky.SpecularCube = mySkyProbe.customBakedTexture;
					mySky.SkyboxCube = mySkyProbe.customBakedTexture;
					mySky.HDRSky = false;
					mySky.HDRSpec = false;

					RegisterAndSyncSkyParameter("masterIntensity", UpdateSkyboxMasterIntensity);
					RegisterAndSyncSkyParameter("diffuseIntensity", UpdateSkyboxDiffuseIntensity);
					RegisterAndSyncSkyParameter("specularIntensity", UpdateSkyboxSpecularIntensity);
					RegisterAndSyncSkyParameter("unityAmbientIntensity", UpdateSkyboxUnityAmbientIntensity);
					RegisterAndSyncSkyParameter("camExposure", UpdateSkyboxCamExposure);
					RegisterAndSyncSkyParameter("skyboxIntensity", UpdateSkyboxSkyIntensity);

					mySky.LinearSpace = !mySkyIsSRGB.val;
					sslc.SetFloatParamValue("skyboxYAngle", 0.0f);
					sslc.skyPopup.onValueChangeHandlers -= UpdateSkyboxName;
					sslc.skyPopup.onValueChangeHandlers += UpdateSkyboxName;
					sslc.skyPopup.currentValueNoCallback = SKY_GAMEOBJECT_NAME;
					sslc.skyManager.BlendToGlobalSky(mySky);
					myHasSetSky = true;
					return true;
				}
			}

			if (myHasSetSky)
			{
				string skyName = sslc.skyName;
				Sky sky = Array.Find(sslc.skies, s => s.name != skyName);
				sslc.skyName = sky.name;
				sslc.skyPopup.onValueChangeHandlers -= UpdateSkyboxName;
				sslc.skyPopup.currentValue = sky.name;
				sslc.skyPopup.onValueChangeHandlers += UpdateSkyboxName;

				UnregisterSkyParameter("masterIntensity", UpdateSkyboxMasterIntensity);
				UnregisterSkyParameter("diffuseIntensity", UpdateSkyboxDiffuseIntensity);
				UnregisterSkyParameter("specularIntensity", UpdateSkyboxSpecularIntensity);
				UnregisterSkyParameter("unityAmbientIntensity", UpdateSkyboxUnityAmbientIntensity);
				UnregisterSkyParameter("camExposure", UpdateSkyboxCamExposure);
				UnregisterSkyParameter("skyboxIntensity", UpdateSkyboxSkyIntensity);

				mySkyProbe = null;
				mySky = null;
				myHasSetSky = false;
			}

			return false;
		}

		private void RegisterAndSyncSkyParameter(string name, JSONStorableFloat.SetFloatCallback callback)
		{
			JSONStorableFloat param = SkyshopLightController.singleton.GetFloatJSONParam(name);
			param.setCallbackFunction += callback;
			callback(param.val);
		}

		private void UnregisterSkyParameter(string name, JSONStorableFloat.SetFloatCallback callback)
		{
			JSONStorableFloat param = SkyshopLightController.singleton.GetFloatJSONParam(name);
			param.setCallbackFunction -= callback;
		}

		private void UpdateSkyboxMasterIntensity(float v)
		{
			if (mySky != null)
				mySky.MasterIntensity = v;
		}

		private void UpdateSkyboxDiffuseIntensity(float v)
		{
			if (mySky != null)
				mySky.DiffIntensity = v;
		}

		private void UpdateSkyboxSpecularIntensity(float v)
		{
			if (mySky != null)
				mySky.SpecIntensity = v;
		}

		private void UpdateSkyboxUnityAmbientIntensity(float v)
		{
			if (mySky != null)
				mySky.DiffIntensityLM = v;
		}

		private void UpdateSkyboxCamExposure(float v)
		{
			if (mySky != null)
				mySky.CamExposure = v;
		}

		private void UpdateSkyboxSkyIntensity(float v)
		{
			if (mySky != null)
				mySky.SkyIntensity = v;
		}

		private void UpdateSkyboxName(string skyName)
		{
			// user changed back to one of the default skies
			mySkyProbe = null;
			mySky = null;
			myHasSetSky = false;
			myUseAsSky.val = false;
		}

		private bool UpdateProbes()
		{
			Transform t = containingAtom?.reParentObject?.Find("object/rescaleObject");
			if (t == null || t.childCount == 0)
				return false;

			ReflectionProbe[] probes = t.GetComponentsInChildren<ReflectionProbe>(true);
			for (int i=0; i<probes.Length; ++i)
				probes[i].intensity = myIntensity.val;
			return true;
		}

		private void UpdatePreview(bool preview)
		{
			if (preview && myPreviewSphere == null)
			{
				myPreviewSphere = GameObject.CreatePrimitive(PrimitiveType.Sphere);
				myPreviewSphere.name = "SkyMagicPreview";
				myPreviewSphere.transform.parent = containingAtom.mainController.transform;
				myPreviewSphere.transform.localPosition = Vector3.zero;
				myPreviewSphere.transform.localRotation = Quaternion.identity;
				myPreviewSphere.transform.localScale = new Vector3(0.25f, 0.25f, 0.25f);
				myPreviewSphere.GetComponent<SphereCollider>().enabled = false;

				Renderer renderer = myPreviewSphere.GetComponent<Renderer>();
				renderer.receiveShadows = false;
				myPreviewMaterial = renderer.material;
				myPreviewMaterial.SetFloat("_Metallic", 1.0f);
				myPreviewMaterial.SetFloat("_Glossiness", 1.0f);
			}
			else if (!preview && myPreviewSphere != null)
			{
				Destroy(myPreviewMaterial);
				Destroy(myPreviewSphere);
				myPreviewSphere = null;
				myPreviewMaterial = null;
			}
		}

		private static void SetupButton(MVRScript script, string label, UnityAction callback, bool rightSide)
		{
			UIDynamicButton button = script.CreateButton(label, rightSide);
			button.button.onClick.AddListener(callback);
		}

		private static JSONStorableBool SetupToggle(MVRScript script, string label, bool defaultValue, bool rightSide)
		{
			JSONStorableBool storable = new JSONStorableBool(label, defaultValue);
			storable.storeType = JSONStorableParam.StoreType.Full;
			script.CreateToggle(storable, rightSide);
			script.RegisterBool(storable);
			return storable;
		}

		private static JSONStorableFloat SetupSliderFloat(MVRScript script, string label, float defaultValue, float minValue, float maxValue, bool rightSide)
		{
			JSONStorableFloat storable = new JSONStorableFloat(label, defaultValue, minValue, maxValue, true, true);
			storable.storeType = JSONStorableParam.StoreType.Full;
			script.CreateSlider(storable, rightSide);
			script.RegisterFloat(storable);
			return storable;
		}

		private static JSONStorableString SetupInfoText(MVRScript script, string text, float height, bool rightSide)
		{
			JSONStorableString storable = new JSONStorableString("Info", text);
			UIDynamic textfield = script.CreateTextField(storable, rightSide);
			textfield.height = height;
			return storable;
		}
    }
}
